/** Sample ALOE component implementation file
 */


/** Include SW API header and SW API utils (skeleton)
*/
#include <phal_sw_api.h>
#include <swapi_utils.h>

/** Include general data type library */
#include "typetools.h"

/** include input/output interfaces and stats.
 * itf_types.h must be included 'before' input.h if it defines e.g. control
 * interface packet structure
*/
#include "itf_types.h"
#include "inputs.h"
#include "outputs.h"
#include "stats.h"

/** declare a buffer to print messages. Default length is defined at
 * swapi_utils.h
 */
static char str[DEF_STR_LEN];


/** Place here your control variables. Let's suppose we want our component to
 * process data in blocks. Our basic variable would indicate the expected length
 * of a block of data. Other examples could be code-rate or constraint length
 * in a decoder, rows and cols in an interleaver, etc.
 */
int blockLength;

/** This is the main processing function. It will be called every time
 * data is available for process. Here you will place your signal processing
 * code.
 *
 * The parameter 'len' indicates the amount of BYTES that have been received
 *
*/
int process_input(int len)
{

    /** In general, we want to make sure we have received the expected
     * amount of data.
     *
     * the function typeSizeArray() returns the expected bytes as a function
     * of a general datatype and expected number of elements. For example, if
     * blockLength=128 and datatype is INTEGER, the function returns 128*4=512.
     * If for example the datatype is BITSTREAM the function will
     * return 128/8=16
     */
    if (len<typeSizeArray(datatype,blockLength)) {
        Log("Sample: Expected length=%d but received only %d bytes\n",typeSizeArray(datatype,blockLength),len);
        /** str is a buffer declared in module.c to use logs
         */ 
        
        return 0;
    }

    /** The most flexible component should be able to work with any data type.
     * Others, however, only have sense working with bits, e.g. CRC or FEC.
     * In the latter case, the only possible input data type will be BITSTREAM.
     * In the former, our module should internally work with either floats or
     * 4-byte integers becuase these are the largest data types. The typetools
     * library offers us a couple of functions to convert an arbitrary
     * datatype to integers/floats to be used by the component. For example,
     * we can conver to integers with the function:
     */
    type2int(input_data,input_temp,blockLength,datatype);

    /**Where the input_data and the input_temp buffers are defined at inputs.h 
     * and stats.h respectively.
     *
     * Note how we convert from a data type which is 'unknown' at compile time.
     * blockLength indicates the amount of samples we want to conver.
     * Alternatively, we have a similar function to convert to floats:
     * type2float(input_data,input_temp,blockLength,datatype);
     */

     /**
      *
      * YOUR SIGNAL PROCESSING FUNCTION GOES HERE
      *
      */
     /** my_proc_function(input_temp,output_temp,blockLength);
      */


    /** Now we convert again to the variable datatype. Note that we could
     * have two 'datatype' variables: one for input and one for output. Thus,
     * our module will be able to convert between types.
     */
    int2type(output_temp,output_data,blockLength,datatype);

    /** The skeleton will read from the output_data buffer the required
     * amount of samples that must be sended.
     */


    /* We return 1 to indicate the processing finished correctly. If we return
     * 0 we will force the module to stop.
     */
    return 1;
}

/** This function returns the amount of bytes the skeleton will send through
 * the output interface (after all input functions finished).
 * Note that again we are using the typetools library to convert from the
 * desired amount of samples to required bytes
 */
int get_output_length() {
    return typeSizeArray(datatype,blockLength);
}

/** Our output process function does nothing. It can be used to make sure
 * we have processed all data, or to generate data if there are not inputs
 */
int output_process(int len) {
    return 1;
}

/** This is the processing function for the control interface. Here, we simply
 * have to validate the parameters and update our internal state variables.
 */
int process_control(int len) {
    if (ctrlpkt.blockLength>INPUT_MAX_DATA) {
        Log("Sample: Invalid blockLength parameter. %d > %d\n",blockLength,INPUT_MAX_DATA);
        
    } else {
        blockLength=ctrlpkt.blockLength;
        Log("Sample: Updated parameter blockLength=%d through control\n",blockLength);
        
    }
    return 1;
}

/** This function is called after all interface processing functions have been
 * called. You can place other state-update code here.
 *
 * In our example, we will use it to read the confBlockLenght statistic (see
 * stats.h) and check if its valid.
 */
int RunCustom()
{
     if (confBlockLength>INPUT_MAX_DATA) {
        Log("Sample: Invalid confBlockLength parameter. %d > %d\n",blockLength,INPUT_MAX_DATA);
        
    } else {
        blockLength=confBlockLength;
        Log("Sample: Updated parameter blockLength=%d through stats\n",blockLength);
        
    }
    return 1;
}

/** This function is called after all interfaces have been initialized.
 * You can place other custom initialization code here.
 */
int InitCustom()
{
    return 1;
}

/** This function is called after initializing parameters and before
 * creating interfaces. It is useful when the number of interfaces or buffers
 * is a function of an initialization parameter.
 */
void ConfigInterfaces()
{
    
}



